;; Generated package description from show-font.el  -*- no-byte-compile: t -*-
(define-package "show-font" "0.1.1.0.20240910.61712" "Show font features in a buffer" '((emacs "28.1")) :commit "6e10b12b5d75785e07e2cc3ba61fdbc754617d22" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "writing" "font") :url "https://github.com/protesilaos/show-font")
